"""Exceptions for GE Kitchen Appliances"""


class GeException(Exception):
    """Base exception for GE Appliance errors"""


class UnknownErdCode(ValueError, GeException):
    """Invalid ERD Code"""


class GeAuthError(RuntimeError, GeException):
    """Failed to authenticate"""



"""Classes to implement GE appliances"""

import logging
from typing import Any, Dict, Optional, Set, TYPE_CHECKING, Union
from slixmpp import JID
from .erd_utils import (
    ERD_DECODERS,
    ERD_ENCODERS,
    ErdCodeType,
    decode_erd_bytes,
    decode_erd_int,
    translate_erd_code,
)
from .erd_constants import ErdApplianceType, ErdCode

if TYPE_CHECKING:
    from .ge_client import GeClient

try:
    import ujson as json
except ImportError:
    import json

_LOGGER = logging.getLogger(__name__)


def _format_request(
        msg_id: Union[int, str], uri: str, method: str, key: Optional[str] = None, value: Optional[str] = None) -> str:
    if method.lower() == 'post':
        post_body = f"<json>{json.dumps({key:value})}</json>"
    else:
        post_body = ""
    message = (
        f"<request><id>{msg_id}</id>"
        f"<method>{method}</method>"
        f"<uri>{uri}</uri>"
        f"{post_body}"
        "</request>"  # [sic.]
    )
    return message


class GeAppliance:
    """Base class shared by all appliances"""

    # Registered encoders and decoders for ERD fields
    _erd_decoders = ERD_DECODERS
    _erd_encoders = ERD_ENCODERS

    def __init__(self, jid: Union[str, JID], client: "GeClient"):
        if not isinstance(jid, JID):
            jid = JID(jid)
        self._available = False
        self._jid = jid
        self._message_id = 0
        self._property_cache = {}  # type: Dict[ErdCodeType, Any]
        self.client = client
        self.initialized = False

    @property
    def jid(self) -> JID:
        return self._jid

    @property
    def known_properties(self) -> Set[ErdCodeType]:
        return set(self._property_cache)

    def send_raw_message(self, mto: JID, mbody: str, mtype: str = 'chat', msg_id: Optional[str] = None):
        """TODO: Use actual xml for this instead of hacking it.  Then again, this is what GE does in the app."""
        if msg_id is None:
            msg_id = self.client.new_id()
        raw_message = (
            f'<message type="{mtype}" from="{self.client.boundjid.bare}" to="{mto}" id="{msg_id}">'
            f'<body>{mbody}</body>'
            f'</message>'
        )
        self.client.send_raw(raw_message)

    def send_request(
            self, method: str, uri: str, key: Optional[str] = None,
            value: Optional[str] = None, message_id: Optional[str] = None):
        """
        Send a pseudo-http request to the appliance
        :param method: str, Usually "GET" or "POST"
        :param uri: str, the "endpoint" for the request, usually of the form "/UUID/erd/{erd_code}"
        :param key: The json key to set in a POST request.  Usually a four-character hex string with leading "0x".
        :param value: The value to set, usually encoded as a hex string without a leading "0x"
        :param message_id:
        """
        if method.lower() != 'post' and (key is not None or value is not None):
            raise RuntimeError('Values can only be set in a POST request')

        if message_id is None:
            message_id = self._message_id
        else:
            self._message_id += 1
            message_id = self._message_id
        message_body = _format_request(message_id, uri, method, key, value)
        self.send_raw_message(self.jid, message_body)

    def request_update(self):
        """Request the appliance send a full state update"""
        self.send_request('GET', '/UUID/cache')

    def set_available(self):
        self._available = True

    def set_unavailable(self):
        self._available = False

    @property
    def available(self) -> bool:
        return self._available and self.initialized

    @property
    def appliance_type(self) -> Optional[ErdApplianceType]:
        return self._property_cache.get(ErdCode.APPLIANCE_TYPE)

    def decode_erd_value(self, erd_code: ErdCodeType, erd_value: str) -> Any:
        """
        Decode and ERD Code raw value into something useful.  If the erd_code is a string that
        cannot be resolved to a known ERD Code, the value will be treated as raw byte string.
        Nonregistered ERD Codes will be translated as ints.

        TODO: Register the numeric fields, and change the default behavior to bytestring

        :param erd_code: ErdCode or str, the ERD Code the value of which we want to decode
        :param erd_value: The raw ERD code value, usually a hex string without leading "0x"
        :return: The decoded value.
        """
        if erd_value == '':
            return None

        erd_code = translate_erd_code(erd_code)

        if isinstance(erd_code, str):
            return decode_erd_bytes(erd_value)

        try:
            return self._erd_decoders[erd_code](erd_value)
        except KeyError:
            return decode_erd_int(erd_value)

    def encode_erd_value(self, erd_code: ErdCodeType, value: Any) -> str:
        """
        Encode an ERD Code value as a hex string.
        Only ERD Codes registered with self.erd_encoders will processed.  Otherwise an error will be returned.

        :param erd_code: ErdCode or str, the ERD Code the value of which we want to decode
        :param value: The value to re-encode
        :return: The encoded value as a hex string
        """
        if value is None:
            return ''

        erd_code = translate_erd_code(erd_code)

        try:
            return self._erd_encoders[erd_code](value)
        except KeyError:
            raise

    def get_erd_value(self, erd_code: ErdCodeType) -> Any:
        """
        Get the value of a property represented by an ERD Code
        :param erd_code: ErdCode or str, The ERD code for the property to get
        :return: The current cached value of that ERD code
        """
        erd_code = translate_erd_code(erd_code)
        return self._property_cache[erd_code]

    def set_erd_value(self, erd_code: ErdCodeType, value: Any):
        """
        Send a new erd value to the appliance
        :param erd_code: The ERD code to update
        :param value: The new value to set
        """
        erd_value = self.encode_erd_value(erd_code, value)
        if isinstance(erd_code, ErdCode):
            raw_erd_code = erd_code.value
        else:
            raw_erd_code = erd_code

        uri = f'/UUID/erd/{raw_erd_code}'
        self.send_request('POST', uri, raw_erd_code, erd_value)

    def update_erd_value(
            self, erd_code: ErdCodeType, erd_value: str) -> bool:
        """
        Setter for ERD code values.

        :param erd_code: ERD code to update
        :param erd_value: The new value to set, as returned by the appliance (usually a hex string)
        :return: Boolean, True if the state changed, False if no value changed
        """
        erd_code = translate_erd_code(erd_code)
        value = self.decode_erd_value(erd_code, erd_value)

        old_value = self._property_cache.get(erd_code)

        try:
            state_changed = ((old_value is None) != (value is None)) or (old_value != value)
        except ValueError:
            _LOGGER.info('Unable to compare new and prior states.')
            state_changed = False

        if state_changed:
            _LOGGER.debug(f'Setting {erd_code} to {value}')
        self._property_cache[erd_code] = value

        return state_changed

    def update_erd_values(self, erd_values: Dict[ErdCodeType, str]) -> Dict[ErdCodeType, Any]:
        """
        Set one or more ERD codes value at once

        :param erd_values: Dictionary of erd codes and their new values as raw hex strings
        :return: dictionary of new states
        """
        state_changes = {
            translate_erd_code(k): self.decode_erd_value(k, v)
            for k, v in erd_values.items()
            if self.update_erd_value(k, v)
        }

        return state_changes

    def __str__(self):
        appliance_type = self.appliance_type
        if appliance_type is None:
            appliance_type = 'Unknown Type'
        return f'{self.__class__.__name__}({self.jid.node}) ({appliance_type})'

    def __format__(self, format_spec):
        return str(self).__format__(format_spec)



"""Conversion functions for ERD values"""

from datetime import timedelta
import logging
from typing import Any, Optional, Set, Tuple, Union
from .erd_constants import (
    ErdApplianceType,
    ErdCode,
    ErdMeasurementUnits,
    ErdAvailableCookMode,
    ErdOvenConfiguration,
    ErdOvenCookMode,
    ErdOvenState,
    OVEN_COOK_MODE_MAP,
)
from .erd_types import (
    AvailableCookMode,
    OvenCookMode,
    OvenCookSetting,
    OvenConfiguration,
)

ErdCodeType = Union[ErdCode, str]

_LOGGER = logging.getLogger(__name__)


def translate_erd_code(erd_code: ErdCodeType) -> ErdCodeType:
    """
    Try to resolve an ERD codes from string to ErdCode if possible.  If an ErdCode
    object is passed in, it will be returned.
    :param erd_code: ErdCode or str
    :return: Either an ErdCode object matching the `erd_code` string, or, if resolution fails,
    the `erd_code` string itself.
    """
    if isinstance(erd_code, ErdCode):
        return erd_code

    try:
        return ErdCode[erd_code]
    except KeyError:
        pass

    try:
        return ErdCode(erd_code)
    except ValueError:
        # raise UnknownErdCode(f"Unable to resolve erd_code '{erd_code}'")
        return erd_code


def decode_erd_int(value: str) -> int:
    """Decode an integer value sent as a hex encoded string."""
    return int(value, 16)


def _encode_erd_int(value: int) -> str:
    """Encode an integer value as a hex string."""
    return value.to_bytes(2, 'big').hex()


def _decode_erd_string(value: str) -> str:
    """
    Decode an string value sent as a hex encoded string.

    TODO: I think the first byte is a checksum.  I need to confirm this so we can have an encoder as well.
    """
    raw_bytes = bytes.fromhex(value)
    raw_bytes = raw_bytes.rstrip(b'\x00')

    return raw_bytes[1:].decode('ascii')


def decode_erd_bytes(value: str) -> bytes:
    """Decode a raw bytes ERD value sent as a hex encoded string."""
    return bytes.fromhex(value)


def _encode_erd_bytes(value: bytes) -> str:
    """Encode a raw bytes ERD value."""
    return value.hex('big')


def _decode_erd_bool(value: str) -> bool:
    return bool(int(value))


def _encode_erd_bool(value: bool) -> str:
    return "01" if value else "00"


def _decode_oven_ranges(value: str) -> Tuple[int, int]:
    raw_bytes = bytes.fromhex(value)
    upper_temp = int.from_bytes(raw_bytes[:2], 'big')
    lower_temp = int.from_bytes(raw_bytes[-2:], 'big')
    return lower_temp, upper_temp


def _decode_appliance_type(value: str) -> ErdApplianceType:
    try:
        return ErdApplianceType(value)
    except ValueError:
        return ErdApplianceType.UNKNOWN


def _decode_measurement_unit(value: str) -> str:
    return ErdMeasurementUnits(int(value))


def _encode_measurement_unit(value: ErdMeasurementUnits) -> str:
    return f'{value.value:02d}'


def _decode_timespan(value: str) -> Optional[timedelta]:
    minutes = decode_erd_int(value)
    if minutes == 65535:
        _LOGGER.debug('Got timespan value of 65535. Treating as None.')
        return None
    return timedelta(minutes=minutes)


def _encode_timespan(value: Optional[timedelta]) -> str:
    if value is None:
        minutes = 65535
    else:
        minutes = value.seconds // 60
    return _encode_erd_int(minutes)


def _decode_cook_modes(value: str) -> Set[AvailableCookMode]:
    if not value:
        return {ErdAvailableCookMode.OVEN_BAKE.value.cook_mode}
    mode_bytes = [int(i) for i in decode_erd_bytes(value)]
    available_modes = {
        mode.value.cook_mode
        for mode in ErdAvailableCookMode
        if mode_bytes[mode.value.byte] & mode.value.mask
    }
    return available_modes


def _decode_oven_state(value: str) -> ErdOvenState:
    """
    See erdCurrentState.smali
    """
    state_code = decode_erd_int(value)
    if 44 <= state_code <= 59:
        return ErdOvenState.OVEN_STATE_SPECIAL_X
    if 42 <= state_code <= 43:
        return ErdOvenState.OVEN_STATE_BAKED_GOODS
    if 40 <= state_code <= 41:
        return ErdOvenState.OVEN_STATE_FROZEN_PIZZA_MULTI
    if 38 <= state_code <= 39:
        return ErdOvenState.OVEN_STATE_FROZEN_SNACKS_MULTI
    if 36 <= state_code <= 37:
        return ErdOvenState.OVEN_STATE_FROZEN_PIZZA
    if 33 <= state_code <= 35:
        return ErdOvenState.OVEN_STATE_FROZEN_SNACKS
    if 1 <= state_code <= 27:
        # These 27 were nicely enumerated in v1.0.3 of the app, though the display logic is more similar to
        # to those above, grouping similar things. See ErdCurrentState.smali for more details.
        return ErdOvenState(state_code)
    return ErdOvenState.STATUS_DASH


def _decode_oven_cook_mode(value: str) -> OvenCookSetting:
    """
    Get the cook mode and temperature.
    TODO: Figure out what the other 10 bytes are for.
        I'm guessing they have to do with two-temp cooking, probes, delayed starts, timers, etc.
    """
    byte_string = decode_erd_bytes(value)
    cook_mode_code = byte_string[0]
    temperature = int.from_bytes(byte_string[1:3], 'big')
    cook_mode = ErdOvenCookMode(cook_mode_code)
    return OvenCookSetting(cook_mode=OVEN_COOK_MODE_MAP[cook_mode], temperature=temperature, raw_bytes=byte_string)


def _encode_oven_cook_mode(cook_setting: OvenCookSetting) -> str:
    """Re-encode a cook mode and temperature
    TODO: Other ten bytes"""
    cook_mode = cook_setting.cook_mode
    cook_mode_code = OVEN_COOK_MODE_MAP.inverse[cook_mode].value
    cook_mode_hex = cook_mode_code.to_bytes(1, 'big').hex()
    temperature_hex = cook_setting.temperature.to_bytes(2, 'big').hex()
    return cook_mode_hex + temperature_hex + ('00' * 10)


def _decode_oven_configuration(value: str) -> OvenConfiguration:
    if not value:
        n = 0
    else:
        n = decode_erd_int(value)

    config = OvenConfiguration(
        has_knob=bool(n & ErdOvenConfiguration.HAS_KNOB.value),
        has_warming_drawer=bool(n & ErdOvenConfiguration.HAS_WARMING_DRAWER.value),
        has_light_bar=bool(n & ErdOvenConfiguration.HAS_LIGHT_BAR.value),
        has_lower_oven=bool(n & ErdOvenConfiguration.HAS_LOWER_OVEN.value),
        has_lower_oven_kitchen_timer=bool(n & ErdOvenConfiguration.HAS_LOWER_OVEN_KITCHEN_TIMER.value),
        raw_value=value,
    )
    return config


# Decoders for non-numeric fields
ERD_DECODERS = {
    # Integers
    ErdCode.CLOCK_FORMAT: decode_erd_int,
    ErdCode.END_TONE: decode_erd_int,
    ErdCode.SOUND_LEVEL: decode_erd_int,
    ErdCode.LOWER_OVEN_DISPLAY_TEMPERATURE: decode_erd_int,
    ErdCode.LOWER_OVEN_PROBE_DISPLAY_TEMP: decode_erd_int,
    ErdCode.LOWER_OVEN_RAW_TEMPERATURE: decode_erd_int,
    ErdCode.LOWER_OVEN_USER_TEMP_OFFSET: decode_erd_int,
    ErdCode.UPPER_OVEN_DISPLAY_TEMPERATURE: decode_erd_int,
    ErdCode.UPPER_OVEN_PROBE_DISPLAY_TEMP: decode_erd_int,
    ErdCode.UPPER_OVEN_RAW_TEMPERATURE: decode_erd_int,
    ErdCode.UPPER_OVEN_USER_TEMP_OFFSET: decode_erd_int,

    # Strings
    ErdCode.MODEL_NUMBER: _decode_erd_string,
    ErdCode.SERIAL_NUMBER: _decode_erd_string,

    # Booleans
    ErdCode.CONVECTION_CONVERSION: _decode_erd_bool,
    ErdCode.HOUR_12_SHUTOFF_ENABLED: _decode_erd_bool,
    ErdCode.SABBATH_MODE: _decode_erd_bool,
    ErdCode.LOWER_OVEN_PROBE_PRESENT: _decode_erd_bool,
    ErdCode.LOWER_OVEN_REMOTE_ENABLED: _decode_erd_bool,
    ErdCode.UPPER_OVEN_PROBE_PRESENT: _decode_erd_bool,
    ErdCode.UPPER_OVEN_REMOTE_ENABLED: _decode_erd_bool,

    # Time spans
    ErdCode.LOWER_OVEN_COOK_TIME_REMAINING: _decode_timespan,
    ErdCode.LOWER_OVEN_DELAY_TIME_REMAINING: _decode_timespan,
    ErdCode.LOWER_OVEN_ELAPSED_COOK_TIME: _decode_timespan,
    ErdCode.LOWER_OVEN_KITCHEN_TIMER: _decode_timespan,
    ErdCode.UPPER_OVEN_COOK_TIME_REMAINING: _decode_timespan,
    ErdCode.UPPER_OVEN_DELAY_TIME_REMAINING: _decode_timespan,
    ErdCode.UPPER_OVEN_ELAPSED_COOK_TIME: _decode_timespan,
    ErdCode.UPPER_OVEN_KITCHEN_TIMER: _decode_timespan,

    # Special handling
    ErdCode.APPLIANCE_TYPE: _decode_appliance_type,
    ErdCode.OVEN_CONFIGURATION: _decode_oven_configuration,
    ErdCode.OVEN_MODE_MIN_MAX_TEMP: _decode_oven_ranges,
    ErdCode.TEMPERATURE_UNIT: _decode_measurement_unit,
    ErdCode.LOWER_OVEN_CURRENT_STATE: _decode_oven_state,
    ErdCode.LOWER_OVEN_AVAILABLE_COOK_MODES: _decode_cook_modes,
    ErdCode.LOWER_OVEN_COOK_MODE: _decode_oven_cook_mode,
    ErdCode.UPPER_OVEN_CURRENT_STATE: _decode_oven_state,
    ErdCode.UPPER_OVEN_AVAILABLE_COOK_MODES: _decode_cook_modes,
    ErdCode.UPPER_OVEN_COOK_MODE: _decode_oven_cook_mode,
}

# Encoders for all fields
ERD_ENCODERS = {
    # Time spans
    ErdCode.LOWER_OVEN_COOK_TIME_REMAINING: _encode_timespan,
    ErdCode.LOWER_OVEN_DELAY_TIME_REMAINING: _encode_timespan,
    ErdCode.LOWER_OVEN_KITCHEN_TIMER: _encode_timespan,
    ErdCode.UPPER_OVEN_COOK_TIME_REMAINING: _encode_timespan,
    ErdCode.UPPER_OVEN_DELAY_TIME_REMAINING: _encode_timespan,
    ErdCode.UPPER_OVEN_KITCHEN_TIMER: _encode_timespan,

    # Booleans
    ErdCode.CONVECTION_CONVERSION: _encode_erd_bool,
    ErdCode.SABBATH_MODE: _encode_erd_bool,

    # Special handling
    ErdCode.LOWER_OVEN_COOK_MODE: _encode_oven_cook_mode,
    ErdCode.UPPER_OVEN_COOK_MODE: _encode_oven_cook_mode,
    ErdCode.TEMPERATURE_UNIT: _encode_measurement_unit,
}



"""Types for structured ERD values"""

from typing import NamedTuple, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .erd_constants import ErdOvenCookMode, ErdOvenState


class AvailableCookMode(NamedTuple):
    """Parsing helper for Available Cook Modes"""
    byte: int
    mask: int
    cook_mode: "ErdOvenCookMode"


class OvenConfiguration(NamedTuple):
    """Cleaner representation of ErdOvenConfiguration"""
    has_knob: bool
    has_warming_drawer: bool
    has_light_bar: bool
    has_lower_oven: bool
    has_lower_oven_kitchen_timer: bool
    raw_value: Optional[str] = None


class OvenCookMode(NamedTuple):
    """Named tuple to represent ErdOvenCookMode for easier formatting later"""
    oven_state: "ErdOvenState"
    delayed: bool = False
    timed: bool = False
    probe: bool = False
    warm: bool = False
    sabbath: bool = False


class OvenCookSetting(NamedTuple):
    """Cleaner representation of ErdOvenCookMode"""
    cook_mode: OvenCookMode
    temperature: int
    raw_bytes: Optional[bytes] = None



"""Constants for handling ERD values"""

import enum
import bidict
from ..erd_types import AvailableCookMode, OvenCookMode


@enum.unique
class ErdApplianceType(enum.Enum):
    UNKNOWN = "FF"
    WATER_HEATER = "00"
    DRYER = "01"
    WASHER = "02"
    FRIDGE = "03"
    MICROWAVE = "04"
    ADVANTIUM = "05"
    DISH_WASHER = "06"
    OVEN = "07"
    ELECTRIC_RANGE = "08"
    GAS_RANGE = "09"
    AIR_CONDITIONER = "0a"
    ELECTRIC_COOKTOP = "0b"
    COOKTOP = "11"
    PIZZA_OVEN = "0c"
    GAS_COOKTOP = "0d"
    SPLIT_AIR_CONDITIONER = "0e"
    HOOD = "0f"
    POE_WATER_FILTER = "10"
    WATER_SOFTENER = "15"
    PORTABLE_AIR_CONDITIONER = "16"
    COMBINATION_WASHER_DRYER = "17"
    ZONELINE = "14"
    DELEVERY_BOX = "12"
    CAFE_COFFEE_MAKER = "1A"


@enum.unique
class ErdMeasurementUnits(enum.Enum):
    IMPERIAL = 0
    METRIC = 1


@enum.unique
class ErdOvenCookMode(enum.Enum):
    """See ErdCookMode.smali"""
    BAKED_GOODS = 60
    BAKETIMEDSHUTOFF_DELAYSTART = 7
    BAKETIMED_TWOTEMP = 5
    BAKETIMED_TWOTEMP_DELAYSTART = 9
    BAKETIMED_WARM = 4
    BAKETIMED_WARM_DELAYSTART = 8
    BAKE_DELAYSTART = 3
    BAKE_NOOPTION = 1
    BAKE_PROBE = 2
    BAKE_PROBE_DELAYSTART = 6
    BAKE_SABBATH = 10
    BROIL_HIGH = 12
    BROIL_LOW = 11
    CONVBAKETIMEDSHUTOFF_DELAYSTART = 24
    CONVBAKETIMED_TWOTEMP = 22
    CONVBAKETIMED_TWOTEMP_DELAYSTART = 26
    CONVBAKETIMED_WARM = 21
    CONVBAKETIMED_WARM_DELAYSTART = 25
    CONVBAKE_DELAYSTART = 20
    CONVBAKE_NOOPTION = 18
    CONVBAKE_PROBE = 19
    CONVBAKE_PROBE_DELAYSTART = 23
    CONVBROILCRISP_NOOPTION = 47
    CONVBROILCRISP_PROBE = 48
    CONVBROIL_HIGH_NOOPTION = 46
    CONVBROIL_LOW_NOOPTION = 45
    CONVMULTIBAKETIMEDSHUTOFF_DELAYSTART = 33
    CONVMULTIBAKETIMED_TWOTEMP = 31
    CONVMULTIBAKETIMED_TWOTEMP_DELAYSTART = 35
    CONVMULTIBAKETIMED_WARM = 30
    CONVMULTIBAKETIMED_WARM_DELAYSTART = 34
    CONVMULTIBAKE_DELAYSTART = 29
    CONVMULTIBAKE_NOOPTION = 27
    CONVMULTIBAKE_PROBE = 28
    CONVMULTIBAKE_PROBE_DELAYSTART = 32
    CONVROASTTIMEDSHUTOFF_DELAYSTART = 42
    CONVROASTTIMED_TWOTEMP = 40
    CONVROASTTIMED_TWOTEMP_DELAYSTART = 44
    CONVROASTTIMED_WARM = 39
    CONVROASTTIMED_WARM_DELAYSTART = 43
    CONVROAST_DELAYSTART = 38
    CONVROAST_NOOPTION = 36
    CONVROAST_PROBE = 37
    CONVROAST_PROBE_DELAYSTART = 41
    CUSTOMSELFCLEAN = 49
    CUSTOMSELFCLEAN_DELAYSTART = 50
    DUALBROIL_HIGH_NOOPTION = 54
    DUALBROIL_LOW_NOOPTION = 53
    FROZEN_PIZZA = 58
    FROZEN_PIZZA_MULTI = 59
    FROZEN_SNACKS = 56
    FROZEN_SNACKS_MULTI = 567
    NOMODE = 0
    PROOF_DELAYSTART = 14
    PROOF_NOOPTION = 13
    STEAMCLEAN = 51
    STEAMCLEAN_DELAYSTART = 52
    WARM_DELAYSTART = 17
    WARM_NOOPTION = 15
    WARM_PROBE = 16


@enum.unique
class ErdAvailableCookMode(enum.Enum):
    """
    Available cooking modes.
    In the XMPP API, they are represented as an index into an array of bytes and a bitmask.
    Thus these take the form (byte: int, mask: int, cook_mode: ErdOvenCookMode).  See ErdAvailableCookMode.smali
    in the Android app.
    """
    OVEN_BAKE = AvailableCookMode(byte=9, mask=2, cook_mode=ErdOvenCookMode.BAKE_NOOPTION)
    OVEN_CONVECTION_BAKE = AvailableCookMode(byte=7, mask=4, cook_mode=ErdOvenCookMode.CONVBAKE_NOOPTION)
    OVEN_CONVECTION_MULTI_BAKE = AvailableCookMode(byte=6, mask=8, cook_mode=ErdOvenCookMode.CONVMULTIBAKE_NOOPTION)
    OVEN_CONVECTION_ROAST = AvailableCookMode(byte=5, mask=16, cook_mode=ErdOvenCookMode.CONVROAST_NOOPTION)
    OVEN_FROZEN_SNACKS = AvailableCookMode(byte=2, mask=1, cook_mode=ErdOvenCookMode.FROZEN_SNACKS)
    OVEN_FROZEN_SNACKS_MULTI = AvailableCookMode(byte=2, mask=2, cook_mode=ErdOvenCookMode.FROZEN_SNACKS_MULTI)
    OVEN_FROZEN_PIZZA = AvailableCookMode(byte=2, mask=4, cook_mode=ErdOvenCookMode.FROZEN_PIZZA)
    OVEN_FROZEN_PIZZA_MULTI = AvailableCookMode(byte=2, mask=8, cook_mode=ErdOvenCookMode.FROZEN_PIZZA_MULTI)
    OVEN_BAKED_GOODS = AvailableCookMode(byte=2, mask=16, cook_mode=ErdOvenCookMode.BAKED_GOODS)


@enum.unique
class ErdOvenConfiguration(enum.Enum):
    """Representation of oven configurations."""
    HAS_KNOB = 1
    HAS_WARMING_DRAWER = 2
    HAS_LIGHT_BAR = 4
    HAS_LOWER_OVEN = 8
    HAS_LOWER_OVEN_KITCHEN_TIMER = 16


@enum.unique
class ErdOvenState(enum.Enum):
    """
    Oven state constants for display purposes.
    These are found in ErdCurrentState.smali.  That they are not aligned with the values for
    ErdCookMode and AvailableCookMode is infuriating.
    """
    BAKE = 5
    BAKE_PREHEAT = 1
    BAKE_TWO_TEMP = 6
    BROIL_HIGH = 14
    BROIL_LOW = 13
    CLEAN_COOL_DOWN = 23
    CLEAN_STAGE1 = 21
    CLEAN_STAGE2 = 22
    CONV_BAKE = 7
    CONV_BAKE_PREHEAT = 2
    CONV_BAKE_TWO_TEMP = 8
    CONV_BROIL_CRISP = 17
    CONV_BROIL_HIGH = 15
    CONV_BROIL_LOW = 16
    CONV_MULTI_BAKE_PREHEAT = 3
    CONV_MULTI_TWO_BAKE = 10
    CONV_MUTLI_BAKE = 9
    CONV_ROAST = 11
    CONV_ROAST2 = 12
    CONV_ROAST_BAKE_PREHEAT = 4
    CUSTOM_CLEAN_STAGE2 = 24
    DELAY = 27
    NO_MODE = 0
    PROOF = 19
    SABBATH = 20
    STEAM_CLEAN_STAGE2 = 25
    STEAM_COOL_DOWN = 26
    WARM = 18
    OVEN_STATE_BAKE = "oven_bake"
    OVEN_STATE_BAKED_GOODS = "oven_state_baked_goods"
    OVEN_STATE_BROIL = "oven_state_broil"
    OVEN_STATE_CONV_BAKE = "oven_state_conv_bake"
    OVEN_STATE_CONV_BAKE_MULTI = "oven_state_conv_bake_multi"
    OVEN_STATE_CONV_BROIL = "oven_state_conv_broil"
    OVEN_STATE_CONV_ROAST = "oven_state_conv_roast"
    OVEN_STATE_DELAY_START = "oven_state_delay_start"
    OVEN_STATE_DUAL_BROIL_HIGH = "oven_state_dual_broil_high"
    OVEN_STATE_DUAL_BROIL_LOW = "oven_state_dual_broil_low"
    OVEN_STATE_FROZEN_PIZZA = "oven_state_frozen_pizza"
    OVEN_STATE_FROZEN_PIZZA_MULTI = "oven_state_frozen_pizza_multi"
    OVEN_STATE_FROZEN_SNACKS = "oven_state_frozen_snacks"
    OVEN_STATE_FROZEN_SNACKS_MULTI = "oven_state_frozen_snacks_multi"
    OVEN_STATE_PROOF = "oven_state_proof"
    OVEN_STATE_SELF_CLEAN = "oven_state_self_clean"
    OVEN_STATE_SPECIAL_X = "oven_state_speical_x"  # [sic.]
    OVEN_STATE_STEAM_START = "oven_state_steam_start"
    OVEN_STATE_WARM = "oven_state_warm"
    STATUS_DASH = "status_dash"


# Translate OVEN_COOK_MODE values into something easier to work with
OVEN_COOK_MODE_MAP = bidict.bidict({
    ErdOvenCookMode.BAKED_GOODS: OvenCookMode(ErdOvenState.OVEN_STATE_BAKED_GOODS, False, False, False),
    ErdOvenCookMode.BAKETIMEDSHUTOFF_DELAYSTART: OvenCookMode(ErdOvenState.BAKE, True, True, False),
    ErdOvenCookMode.BAKETIMED_TWOTEMP: OvenCookMode(ErdOvenState.BAKE_TWO_TEMP, False, True, False),
    ErdOvenCookMode.BAKETIMED_TWOTEMP_DELAYSTART: OvenCookMode(ErdOvenState.BAKE_TWO_TEMP, True, True, False),
    ErdOvenCookMode.BAKETIMED_WARM: OvenCookMode(ErdOvenState.WARM, False, True, False),
    ErdOvenCookMode.BAKETIMED_WARM_DELAYSTART: OvenCookMode(ErdOvenState.WARM, True, True, False),
    ErdOvenCookMode.BAKE_DELAYSTART: OvenCookMode(ErdOvenState.BAKE, True, False, False),
    ErdOvenCookMode.BAKE_NOOPTION: OvenCookMode(ErdOvenState.BAKE, False, False, False),
    ErdOvenCookMode.BAKE_PROBE: OvenCookMode(ErdOvenState.BAKE, False, False, True),
    ErdOvenCookMode.BAKE_PROBE_DELAYSTART: OvenCookMode(ErdOvenState.BAKE, True, False, True),
    ErdOvenCookMode.BAKE_SABBATH: OvenCookMode(ErdOvenState.BAKE, False, False, False, False, True),
    ErdOvenCookMode.BROIL_HIGH: OvenCookMode(ErdOvenState.BROIL_HIGH, False, False, False),
    ErdOvenCookMode.BROIL_LOW: OvenCookMode(ErdOvenState.BROIL_LOW, False, False, False),
    ErdOvenCookMode.CONVBAKETIMEDSHUTOFF_DELAYSTART: OvenCookMode(ErdOvenState.CONV_BAKE, True, True, False),
    ErdOvenCookMode.CONVBAKETIMED_TWOTEMP: OvenCookMode(ErdOvenState.CONV_BAKE_TWO_TEMP, False, True, False),
    ErdOvenCookMode.CONVBAKETIMED_TWOTEMP_DELAYSTART: OvenCookMode(ErdOvenState.CONV_BAKE_TWO_TEMP, True, True, False),
    ErdOvenCookMode.CONVBAKETIMED_WARM: OvenCookMode(ErdOvenState.CONV_BAKE, False, True, False, True),
    ErdOvenCookMode.CONVBAKETIMED_WARM_DELAYSTART: OvenCookMode(ErdOvenState.CONV_BAKE, True, True, False, True),
    ErdOvenCookMode.CONVBAKE_DELAYSTART: OvenCookMode(ErdOvenState.CONV_BAKE, True, False, False),
    ErdOvenCookMode.CONVBAKE_NOOPTION: OvenCookMode(ErdOvenState.CONV_BAKE, False, False, False),
    ErdOvenCookMode.CONVBAKE_PROBE: OvenCookMode(ErdOvenState.CONV_BAKE, False, False, True),
    ErdOvenCookMode.CONVBAKE_PROBE_DELAYSTART: OvenCookMode(ErdOvenState.CONV_BAKE, True, False, True),
    ErdOvenCookMode.CONVBROILCRISP_NOOPTION: OvenCookMode(ErdOvenState.CONV_BROIL_CRISP, False, False, False),
    ErdOvenCookMode.CONVBROILCRISP_PROBE: OvenCookMode(ErdOvenState.CONV_BROIL_CRISP, False, False, True),
    ErdOvenCookMode.CONVBROIL_HIGH_NOOPTION: OvenCookMode(ErdOvenState.CONV_BROIL_HIGH, False, False, False),
    ErdOvenCookMode.CONVBROIL_LOW_NOOPTION: OvenCookMode(ErdOvenState.CONV_BROIL_LOW, False, False, False),
    ErdOvenCookMode.CONVMULTIBAKETIMEDSHUTOFF_DELAYSTART: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, True, True, False),
    ErdOvenCookMode.CONVMULTIBAKETIMED_TWOTEMP: OvenCookMode(ErdOvenState.CONV_MULTI_TWO_BAKE, False, True, False),
    ErdOvenCookMode.CONVMULTIBAKETIMED_TWOTEMP_DELAYSTART: OvenCookMode(ErdOvenState.CONV_MULTI_TWO_BAKE, True, True, False),
    ErdOvenCookMode.CONVMULTIBAKETIMED_WARM: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, False, True, False, True),
    ErdOvenCookMode.CONVMULTIBAKETIMED_WARM_DELAYSTART: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, True, True, False, True),
    ErdOvenCookMode.CONVMULTIBAKE_DELAYSTART: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, True, False, False),
    ErdOvenCookMode.CONVMULTIBAKE_NOOPTION: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, False, False, False),
    ErdOvenCookMode.CONVMULTIBAKE_PROBE: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, False, False, True),
    ErdOvenCookMode.CONVMULTIBAKE_PROBE_DELAYSTART: OvenCookMode(ErdOvenState.CONV_MUTLI_BAKE, True, False, True),
    ErdOvenCookMode.CONVROASTTIMEDSHUTOFF_DELAYSTART: OvenCookMode(ErdOvenState.CONV_ROAST, True, True, False),
    ErdOvenCookMode.CONVROASTTIMED_TWOTEMP: OvenCookMode(ErdOvenState.CONV_ROAST2, False, True, False),
    ErdOvenCookMode.CONVROASTTIMED_TWOTEMP_DELAYSTART: OvenCookMode(ErdOvenState.CONV_ROAST2, True, True, False),
    ErdOvenCookMode.CONVROASTTIMED_WARM: OvenCookMode(ErdOvenState.CONV_ROAST, False, True, False, True),
    ErdOvenCookMode.CONVROASTTIMED_WARM_DELAYSTART: OvenCookMode(ErdOvenState.CONV_ROAST, True, True, False, True),
    ErdOvenCookMode.CONVROAST_DELAYSTART: OvenCookMode(ErdOvenState.CONV_ROAST, True, False, False),
    ErdOvenCookMode.CONVROAST_NOOPTION: OvenCookMode(ErdOvenState.CONV_ROAST, False, False, False),
    ErdOvenCookMode.CONVROAST_PROBE: OvenCookMode(ErdOvenState.CONV_ROAST, False, False, True),
    ErdOvenCookMode.CONVROAST_PROBE_DELAYSTART: OvenCookMode(ErdOvenState.CONV_ROAST, True, False, True),
    ErdOvenCookMode.CUSTOMSELFCLEAN: OvenCookMode(ErdOvenState.CUSTOM_CLEAN_STAGE2, False, False, False),
    ErdOvenCookMode.CUSTOMSELFCLEAN_DELAYSTART: OvenCookMode(ErdOvenState.CUSTOM_CLEAN_STAGE2, True, False, False),
    ErdOvenCookMode.DUALBROIL_HIGH_NOOPTION: OvenCookMode(ErdOvenState.OVEN_STATE_DUAL_BROIL_HIGH, False, False, False),
    ErdOvenCookMode.DUALBROIL_LOW_NOOPTION: OvenCookMode(ErdOvenState.OVEN_STATE_DUAL_BROIL_LOW, False, False, False),
    ErdOvenCookMode.FROZEN_PIZZA: OvenCookMode(ErdOvenState.OVEN_STATE_FROZEN_PIZZA, False, False, False),
    ErdOvenCookMode.FROZEN_PIZZA_MULTI: OvenCookMode(ErdOvenState.OVEN_STATE_FROZEN_PIZZA_MULTI, False, False, False),
    ErdOvenCookMode.FROZEN_SNACKS: OvenCookMode(ErdOvenState.OVEN_STATE_FROZEN_SNACKS, False, False, False),
    ErdOvenCookMode.FROZEN_SNACKS_MULTI: OvenCookMode(ErdOvenState.OVEN_STATE_FROZEN_SNACKS_MULTI, False, False, False),
    ErdOvenCookMode.NOMODE: OvenCookMode(ErdOvenState.NO_MODE, False, False, False),
    ErdOvenCookMode.PROOF_DELAYSTART: OvenCookMode(ErdOvenState.PROOF, True, False, False),
    ErdOvenCookMode.PROOF_NOOPTION: OvenCookMode(ErdOvenState.PROOF, False, False, False),
    ErdOvenCookMode.STEAMCLEAN: OvenCookMode(ErdOvenState.STEAM_CLEAN_STAGE2, False, False, False),
    ErdOvenCookMode.STEAMCLEAN_DELAYSTART: OvenCookMode(ErdOvenState.STEAM_CLEAN_STAGE2, True, False, False),
    ErdOvenCookMode.WARM_DELAYSTART: OvenCookMode(ErdOvenState.WARM, True, False, False),
    ErdOvenCookMode.WARM_NOOPTION: OvenCookMode(ErdOvenState.WARM, False, False, False),
    ErdOvenCookMode.WARM_PROBE: OvenCookMode(ErdOvenState.WARM, False, False, True),
})



import asyncio
import logging
import slixmpp
import socket
import ssl
from lxml import etree
from .const import (
    EVENT_ADD_APPLIANCE,
    EVENT_APPLIANCE_STATE_CHANGE,
    EVENT_APPLIANCE_INITIAL_UPDATE,
)
from .erd_constants import ErdCode
from .erd_utils import ErdCodeType
from .ge_appliance import GeAppliance
from typing import Any, Dict, Optional, Set, Tuple

try:
    import ujson as json
except ImportError:
    import json

_LOGGER = logging.getLogger(__name__)


def _first_or_none(lst: list) -> Any:
    try:
        return lst[0]
    except IndexError:
        return None


set_timer = False


class GeClient(slixmpp.ClientXMPP):
    def __init__(self, xmpp_credentials: Dict, event_loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs):
        # If we have our own even loop, use that
        self._loop = event_loop
        self._xmpp_credentials = xmpp_credentials

        super().__init__(xmpp_credentials['jid'], xmpp_credentials['password'], **kwargs)
        self.register_plugin('xep_0030')  # Service Discovery
        self.register_plugin('xep_0199')  # XMPP Ping

        self.add_event_handler('message', self.on_message)
        self.add_event_handler('session_start', self.on_start)
        self.add_event_handler('presence_available', self.on_presence_available)
        self.add_event_handler('presence_unavailable', self.on_presence_unavailable)
        self.add_event_handler(EVENT_APPLIANCE_STATE_CHANGE, self.maybe_trigger_appliance_init_event)
        self.ssl_context.verify_mode = ssl.CERT_NONE
        self._appliances = {}  # type: Dict[str, GeAppliance]
        self._initial_update_ids = set()  # type: Set[str]

    def connect(self, address: Optional[Tuple[str, int]] = None, use_ssl: bool = False,
                force_starttls: bool = True, disable_starttls: bool = False):
        """Connect to the XMPP server."""
        if address is None:
            address = (self._xmpp_credentials['address'], self._xmpp_credentials['port'])
        super().connect(
            address=address,
            use_ssl=use_ssl,
            force_starttls=force_starttls,
            disable_starttls=disable_starttls
        )

    @property
    def appliances(self) -> Dict[str, GeAppliance]:
        return self._appliances

    async def add_appliance(self, jid: str):
        """Add an appliance to the registry and request an update."""
        if jid in self.appliances:
            raise RuntimeError('Trying to add duplicate appliance')
        new_appliance = GeAppliance(jid, self)

        _LOGGER.info(f'Adding appliance {jid}')
        self.appliances[jid] = new_appliance
        self.event(EVENT_ADD_APPLIANCE, new_appliance)
        new_appliance.request_update()

    async def on_presence_available(self, evt: slixmpp.ElementBase):
        """Perform actions when notified of an available JID."""
        jid = slixmpp.JID(evt['from']).bare
        if jid == self.boundjid.bare:
            return
        try:
            self.appliances[jid].set_available()
            _LOGGER.debug(f'Appliance {jid} marked available')
        except KeyError:
            await self.add_appliance(jid)

    async def on_presence_unavailable(self, evt):
        """When appliance is no longer available, mark it as such."""
        jid = slixmpp.JID(evt['from']).bare
        if jid == self.boundjid.bare:
            return
        try:
            self.appliances[jid].set_unavailable()
            _LOGGER.debug(f'Appliance {jid} marked unavailable')
        except KeyError:
            pass

    async def maybe_trigger_appliance_init_event(self, data: Tuple[GeAppliance, Dict[ErdCodeType, Any]]):
        """
        Trigger the appliance_got_type event if appropriate

        :param data: GeAppliance updated and the updates
        """
        appliance, state_changes = data
        if ErdCode.APPLIANCE_TYPE in state_changes and not appliance.initialized:
            _LOGGER.debug(f'Got initial appliance type for {appliance:s}')
            appliance.initialized = True
            self.event(EVENT_APPLIANCE_INITIAL_UPDATE, appliance)

    async def on_message(self, event):
        """Handle incoming messages."""
        global set_timer
        msg = str(event)
        msg_from = slixmpp.JID(event['from']).bare
        try:
            message_data = self.extract_message_json(msg)
        except ValueError:
            _LOGGER.info(f"From: {msg_from}: Not a GE message")
            return
        try:
            appliance = self.appliances[msg_from]
            state_changes = appliance.update_erd_values(message_data)
            if state_changes:
                self.event(EVENT_APPLIANCE_STATE_CHANGE, [appliance, state_changes])
        except KeyError:
            _LOGGER.warning('Received json message from unregistered appliance')

    async def on_start(self, event):
        _LOGGER.debug('Starting session: ' + str(event))
        self.send_presence('available')

    def complete_jid(self, jid) -> str:
        """Make a complete jid from a username"""
        if '@' in jid:
            return jid
        return f'{jid}@{self.boundjid.host}'

    @staticmethod
    def extract_message_json(message: str) -> Dict:
        """The appliances send messages that don't play nice with slixmpp, so let's do this."""
        etr = etree.XML(message)
        json_elem = _first_or_none(etr.xpath('//json'))
        if json_elem is None:
            raise ValueError('Not a GE appliance message')

        data = _first_or_none(json_elem.xpath('text()'))
        data = json.loads(data)
        return data

    async def process_in_running_loop(self, timeout: Optional[int] = None):
        """Process all the available XMPP events (receiving or sending data on the
        socket(s), calling various registered callbacks, calling expired
        timers, handling signal events, etc).  If timeout is None, this
        coroutine will run forever or until disconnected. If timeout is a number,
        this function will return after the given time in seconds."""

        if timeout is None:
            await asyncio.ensure_future(self.disconnected, loop=self.loop)
        else:
            await asyncio.wait([self.disconnected], loop=self.loop, timeout=timeout)

    async def _connect_routine(self):
        """
        Override the _connect_routine method from xmlstream.py.
        This is to address the bug corrected in open PR https://github.com/poezio/slixmpp/pull/19.
        """
        self.event_when_connected = "connected"

        if self.connect_loop_wait > 0:
            self.event('reconnect_delay', self.connect_loop_wait)
            await asyncio.sleep(self.connect_loop_wait, loop=self.loop)

        record = await self.pick_dns_answer(self.default_domain)
        if record is not None:
            host, address, dns_port = record
            port = self.address[1] if self.address[1] else dns_port
            self.address = (address, port)
            self._service_name = host
        else:
            # No DNS records left, stop iterating
            # and try (host, port) as a last resort
            self.dns_answers = None

        if self.use_ssl:
            ssl_context = self.get_ssl_context()
        else:
            ssl_context = None

        if self._current_connection_attempt is None:
            return
        try:
            await self.loop.create_connection(
                lambda: self, self.address[0], self.address[1], ssl=ssl_context,
                server_hostname=self.default_domain if self.use_ssl else None
            )
            self.connect_loop_wait = 0
        except socket.gaierror:
            self.event('connection_failed',
                       'No DNS record available for %s' % self.default_domain)
        except OSError as e:
            _LOGGER.debug('Connection failed: %s', e)
            self.event("connection_failed", e)
            if self._current_connection_attempt is None:
                return
            self.connect_loop_wait = self.connect_loop_wait * 2 + 1
            self._current_connection_attempt = asyncio.ensure_future(
                self._connect_routine(), loop=self.loop,
            )



"""GE Constants"""

from .erd_codes import ErdCode
from .erd_constants import (
    ErdApplianceType,
    ErdAvailableCookMode,
    ErdMeasurementUnits,
    ErdOvenConfiguration,
    ErdOvenCookMode,
    ErdOvenState,
    OVEN_COOK_MODE_MAP,
)



"""GE Kitchen Appliances SDK"""

__version__ = "0.1.6"


from .async_login_flow import (
    AbstractAuth,
    async_do_full_login_flow,
    async_get_ge_token,
    async_get_mobile_device_token,
    async_get_oauth2_token,
    async_get_xmpp_credentials,
)
from .const import *
from .erd_constants import *
from .erd_types import AvailableCookMode, OvenConfiguration, OvenCookSetting, OvenCookMode
from .erd_utils import ERD_DECODERS, ERD_ENCODERS, ErdCodeType, translate_erd_code
from .exc import *
from .ge_appliance import GeAppliance
from .ge_client import GeClient
from .login_flow import (
    do_full_login_flow,
    get_ge_token,
    get_mobile_device_token,
    get_oauth2_token,
    get_xmpp_credentials,
)




"""Constants"""

# OAuth2 credentials
OAUTH2_APP_ID = 'com.ge.kitchen.wca.prd.android'
OAUTH2_CLIENT_ID = '564c31616c4f7474434b307435412b4d2f6e7672'
OAUTH2_CLIENT_SECRET = '6476512b5246446d452f697154444941387052645938466e5671746e5847593d'
OAUTH2_REDIRECT_URI = 'brillion.4e617a766474657344444e562b5935566e51324a://oauth/redirect'

LOGIN_URL = 'https://accounts.brillion.geappliances.com'
API_URL = 'https://api.brillion.geappliances.com'
SECURE_URL = 'https://secure.brillion.geappliances.com'

EVENT_APPLIANCE_STATE_CHANGE = 'appliance_state_change'
EVENT_ADD_APPLIANCE = 'add_appliance'
EVENT_APPLIANCE_INITIAL_UPDATE = 'appliance_got_type'
